
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include "tools.h"

void rgb32_blit_scanline( unsigned char *output, unsigned char *input, int width )
{
    unsigned int *out = (unsigned int *) output;
    unsigned int *in = (unsigned int *) input;
    while( width-- ) {
        *out++ = *in++;
    }
}

void rgb32_blit_colour_scanline( unsigned char *output, unsigned int colour, int width )
{
    unsigned int *ptr = (unsigned int *) output;
    while( width-- ) {
        *ptr++ = colour;
    }
}

void rgb32_blit_colour( unsigned char *output, unsigned int colour, int width, int height, int stride )
{
    while( height-- ) {
        rgb32_blit_colour_scanline( output, colour, width );
        output += stride;
    }
}

void rgb32_blit_colour_vertical( unsigned char *output, unsigned int colour, int height, int stride )
{
    unsigned int *ptr = (unsigned int *) output;
    stride /= 4;
    while( height-- ) {
        *ptr = colour;
        ptr += stride;
    }
}

void rgb24_to_rgb32_scanline( unsigned char *output, unsigned char *input, int width )
{
    while( width-- ) {
        output[ 0 ] = 0;
        output[ 1 ] = input[ 2 ];
        output[ 2 ] = input[ 1 ];
        output[ 3 ] = input[ 0 ];
        output += 4;
        input += 3;
    }
}

void argb32_to_rgba32_scanline( unsigned char *output, unsigned char *input, int width )
{
    while( width-- ) {
        output[ 0 ] = input[ 1 ];
        output[ 1 ] = input[ 2 ];
        output[ 2 ] = input[ 3 ];
        output[ 3 ] = input[ 0 ];
        output += 4;
        input += 4;
    }
}

void rgba32_to_argb32_scanline( unsigned char *output, unsigned char *input, int width )
{
    while( width-- ) {
        output[ 0 ] = input[ 2 ]; // blue
        output[ 1 ] = input[ 1 ]; // green
        output[ 2 ] = input[ 0 ]; // red
        output[ 3 ] = input[ 3 ]; // alpha
        output += 4;
        input += 4;
    }
}

void rgb32_to_rgb16_scanline( unsigned char *output, unsigned char *input, int width )
{
    unsigned int *ptr = (unsigned int *) input;
    unsigned short *out = (unsigned short *) output;
    while( width-- ) {
        int cur = *ptr, r, g, b;
        r = ((cur & 0x00ff0000) >> 16) >> 3;
        g = ((cur & 0x0000ff00) >>  8) >> 2;
        b = ((cur & 0x000000ff)      ) >> 3;
        *out = (r << 11) | (g << 5) | b;
        ptr++;
        out++;
    }
}

int tga_to_rgb32( unsigned char *dest, const char *filename, int width, int height, int stride )
{
    unsigned char headerSize;
    unsigned short w, h;
    unsigned char bpp;
    int x, y;

    int fd = open( filename, O_RDONLY );
    if( fd < 0 ) return 0;

    if( read( fd, &headerSize, 1 ) != 1 ) {
        close( fd );
        return 0;
    }

    lseek( fd, 0xC, SEEK_SET );
    if( read( fd, &w, 2 ) != 2 || read( fd, &h, 2 ) != 2 ) {
        close( fd );
        return 0;
    }
    read( fd, &bpp, 1 );

    /* Clip image width and height to buffer width and height. */
    if( w < width ) width = w;
    if( h < height ) height = h;
    bpp /= 8;

    lseek( fd, 18 + headerSize, SEEK_SET );

    /* TGA files are backwards. */
    dest += stride * ( height - 1 );

    for( y = h - height; y; y-- ) {
        for( x = w; x; x-- ) {
            unsigned char pixels[ 4 ];
            read( fd, pixels, bpp );
        }
    }
    for( y = height; y; y-- ) {
        unsigned char *scanline = dest;
        for( x = width; x; x-- ) {
            unsigned char pixels[ 4 ];
            pixels[ 3 ] = 0xff;
            read( fd, pixels, bpp );
            scanline[ 0 ] = pixels[ 0 ]; /* Blue */
            scanline[ 1 ] = pixels[ 1 ]; /* Green */
            scanline[ 2 ] = pixels[ 2 ]; /* Red */
            scanline[ 3 ] = pixels[ 3 ]; /* Alpha */
            scanline += 4;
        }
        for( x = w - width; x; x-- ) {
            unsigned char pixels[ 4 ];
            read( fd, pixels, bpp );
        }
        dest -= stride;
    }
    close( fd );
    return 1;
}

void a8_subpix_blit_scanline( unsigned char *output, unsigned char *input,
                              int startpos, int width )
{
    int pos = 0xffff - (startpos & 0xffff);
    int prev = input[ 0 ];
    int tmp;
    int x;

    for( x = 0; x < width; x++ ) {
        tmp = ( ( prev * pos ) + ( input[ x ] * ( 0xffff - pos ) ) ) / 65535;
        tmp += output[ x ];
        output[ x ] = (tmp > 255) ? 255 : tmp;
        prev = input[ x ];
    }
    /*
    tmp = ( prev * pos ) / 65535;
    tmp += output[ width ];
    output[ width ] = (tmp > 255) ? 255 : tmp;
    */
}

void rgb32_subpix_blit_scanline( unsigned char *output, unsigned char *input,
                                 unsigned char lastr, unsigned char lastg, unsigned char lastb,
                                 int startpos, int width )
{
    int pos = 0xffff - (startpos & 0xffff);
    int prevr = lastr;
    int prevg = lastg;
    int prevb = lastb;
    int tmp;
    int x;

    for( x = 0; x < width*4; x += 4 ) {
        output[ x ] = ( ( prevb * pos ) + ( input[ x ] * ( 0xffff - pos ) ) ) >> 16;
        prevb = input[ x ];

        output[ x + 1 ] = ( ( prevg * pos ) + ( input[ x + 1 ] * ( 0xffff - pos ) ) ) >> 16;
        prevg = input[ x + 1 ];

        output[ x + 2 ] = ( ( prevr * pos ) + ( input[ x + 2 ] * ( 0xffff - pos ) ) ) >> 16;
        prevr = input[ x + 2 ];
    }
}

