
#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include "input.h"
#include "pnginput.h"
#include "xoutput.h"
#include "tools.h"
#include "leetft.h"
#include "speedy.h"
#include "rtctimer.h"
#include "vgasync.h"

static int timediff( struct timeval *large, struct timeval *small )
{
    return (   ( ( large->tv_sec * 1000 * 1000 ) + large->tv_usec )
             - ( ( small->tv_sec * 1000 * 1000 ) + small->tv_usec ) );
}

static double intensity_to_voltage_srgb( double val )
{
    /* Handle invalid values before doing a gamma transform. */
    if( val < 0.0 ) return 0.0;
    if( val > 1.0 ) return 1.0;

    /* sRGB uses an exponent of 2.4 in its nonstandard curve, but it
     *      * still advertises a display gamma of 2.2.
     *           */
    if( val <= 0.0031308 ) {
        return 12.92 * val;
    }
    return ( ( 1.055 * pow( val, 1.0 / 2.4 ) ) - 0.055 );
}

static int crawlpage = 0;
static int crawlpos = 0;
static pnginput_t *curpng = 0;

int load_next_png( void )
{
    char filename[ 1024 ];

    if( curpng ) pnginput_delete( curpng );
    sprintf( filename, "/tmp/crawl%08d.png", crawlpage++ );
    curpng = pnginput_new( filename );
    if( !curpng ) {
        crawlpage = 0;
        sprintf( filename, "/tmp/crawl%08d.png", crawlpage++ );
        curpng = pnginput_new( filename );
        if( !curpng ) {
            fprintf( stderr, "viewcrawl: Can't find %s: nothing to draw.\n", filename );
            return 0;
        }
    }
    crawlpos = 0;

    return 1;
}

int fill_next_page( unsigned char *data, int width, int height, int stride )
{
    int curpos = 0;
    int startpos = 0;
    int i;
    int done = 0;

    while( curpos < width ) {
        int curwidth = pnginput_get_width( curpng ) - crawlpos;
        int loadnext = 1;

        startpos = crawlpos;
        if( curwidth >= ( width - curpos ) ) {
            curwidth = ( width - curpos );
            crawlpos += curwidth;
            loadnext = 0;
            done = 1;
        }

        for( i = 0; i < height; i++ ) {
            unsigned char *curscanline = data + (i * stride) + (curpos * 4);

            if( i >= pnginput_get_height( curpng ) ) {
                blit_colour_packed4444_scanline( curscanline, curwidth, 0, 0, 0, 0 );
            } else {
                rgba32_to_argb32_scanline( curscanline,
                                           pnginput_get_scanline( curpng, i ) + (startpos * 4),
                                           curwidth );
            }
        }
        curpos += curwidth;

        if( loadnext ) {
            if( !load_next_png() ) {
                return 0;
            }
        }
    }
    return 1;
}

int main( int argc, char **argv )
{
    output_api_t *output = xoutput_get_output();
    input_t *in = input_new();
    int width, height, stride;
    unsigned char *page1;
    unsigned char *page2;
    struct timeval lastblit;
    struct timeval curtime;
    int pos = 0;
    int pagewidth, pageheight, pagestride;
    int i;
    int usecperframe;
    int refreshtime;
    int blittime = 0;

    int accel_step = 10000;
    int accel_max;
    int accel = 0;
    int paused = 1;
    rtctimer_t *rtctimer;
    int error = 0;
    int use_vgasync = 1;

    int pixels_per_second = 200;

    setup_speedy_calls( 1 );

    if( !load_next_png() ) {
        return 1;
    }

    if( !vgasync_init( 1 ) ) {
        fprintf( stderr, "viewcrawl: Can't open VGA port, must be root.\n" );
        fprintf( stderr, "viewcrawl: Assuming 85hz refresh.\n" );
        refreshtime = 11765;
        use_vgasync = 0;
    }
    if( !set_realtime_priority( 0 ) ) {
        fprintf( stderr, "viewcrawl: Can't set realtime priority, must be root.  Accuracy will suffer.\n" );
    }
    rtctimer = rtctimer_new( 1 );
    if( !rtctimer ) {
        fprintf( stderr, "viewcrawl: Can't open /dev/rtc needed for timing, defaulting to usleep accuracy (10ms).\n" );
    } else {
        if( !rtctimer_set_interval( rtctimer, 1024 ) ) {
            fprintf( stderr, "viewcrawl: Can't get millisecond accuracy from /dev/rtc, you must be root or change max-user-freq in /proc.\n" );
            rtctimer_delete( rtctimer );
            rtctimer = 0;
        } else {
            rtctimer_start_clock( rtctimer );
        }
    }

    width = 800;
    height = 60;
    output->init( width, height );
    stride = output->get_output_stride();

    pagewidth = width;
    pageheight = height;
    pagestride = pagewidth * 4;

    page1 = (unsigned char *) malloc( pagestride * height );
    page2 = (unsigned char *) malloc( pagestride * height );
    if( !page1 || !page2 ) {
        fprintf( stderr, "viewcrawl: Can't allocte memory.\n" );
        return 1;
    }

    if( !fill_next_page( page1, pagewidth, pageheight, pagestride ) ) {
        fprintf( stderr, "viewcrawl: Can't open next crawl file, exiting.\n" );
    }
    if( !fill_next_page( page2, pagewidth, pageheight, pagestride ) ) {
        fprintf( stderr, "viewcrawl: Can't open next crawl file, exiting.\n" );
    }

    output->lock_output_buffer();
    memset( output->get_output_buffer(), 0, stride * height );
    output->unlock_output_buffer();
    output->show_frame();

    /* Try to measure refresh rate. */
    if( use_vgasync ) {
        vgasync_spin_until_end_of_next_refresh();
        vgasync_spin_until_end_of_next_refresh();
        gettimeofday( &lastblit, 0 );
        vgasync_spin_until_end_of_next_refresh();
        gettimeofday( &curtime, 0 );
        refreshtime = timediff( &curtime, &lastblit );
        fprintf( stderr, "viewcrawl: Measured refresh rate: %.2fhz.\n",
                 (1.0 / ((double)refreshtime))*1000.0 * 1000.0 );
        vgasync_spin_until_inside_refresh();
        gettimeofday( &curtime, 0 );
        lastblit = curtime;
    } else {
        gettimeofday( &lastblit, 0 );
    }

    // Blit a frame every 1.5 refresh times, for now.
    usecperframe = refreshtime + (refreshtime/2);

    // Start with 200 pixels per second.
    accel_max = ((pixels_per_second * usecperframe) / (1000*1000)) << 16;

    for(;;) {
        struct timeval beforeblit;
        struct timeval afterblit;

        output->poll_events( in );
        if( input_quit_requested( in ) ) break;
        if( input_pause_requested( in ) ) paused = !paused;
        input_next_frame( in );

        if( !paused ) {
            pos += accel;
            if( accel < accel_max ) {
                accel += accel_step;
            }
        } else {
            pos += accel;
            if( accel > 0 ) {
                accel -= accel_step;
            }
        }

        if( pos > (pagewidth << 16) ) {
            unsigned char *tmpp = page1;
            page1 = page2;
            page2 = tmpp;
            if( !fill_next_page( page2, pagewidth, pageheight, pagestride ) ) {
                fprintf( stderr, "viewcrawl: Can't open next crawl file, exiting.\n" );
                return 1;
            }
            pos -= (pagewidth << 16 );
        }

        /* Render */
        output->lock_output_buffer();
        for( i = 0; i < height; i++ ) {
            unsigned char *curscanline = output->get_output_buffer() + (stride * i);
            unsigned char *curpage1 = page1 + (i * pagestride);
            unsigned char *curpage2 = page2 + (i * pagestride);
            int pixpos = (pos >> 16);
            rgb32_subpix_blit_scanline( curscanline,
                                        curpage1 + (pixpos * 4),
                                        0, 0, 0,
                                        pos & 0xffff,
                                        pagewidth - pixpos );
            if( pos ) {
                unsigned char lastr = curpage1[ (pagewidth*4) - 2 ];
                unsigned char lastg = curpage1[ (pagewidth*4) - 3 ];
                unsigned char lastb = curpage1[ (pagewidth*4) - 4 ];
                rgb32_subpix_blit_scanline( curscanline + ((width - pixpos) * 4),
                                            curpage2, lastr, lastg, lastb,
                                            (pos & 0xffff), pixpos );
            }
        }
        output->unlock_output_buffer();

        /* Wait until blit time. */
        for(;;) {
            gettimeofday( &curtime, 0 );
            if( timediff( &curtime, &lastblit ) > ( usecperframe - blittime + error ) ) {
                break;
            }
            if( rtctimer ) {
                rtctimer_next_tick( rtctimer );
            } else {
                /* Try to usleep for 1ms. */
                usleep( 1000 );
            }
        }
        gettimeofday( &beforeblit, 0 );
        output->show_frame();
        gettimeofday( &afterblit, 0 );
        blittime = timediff( &afterblit, &beforeblit );
        error += usecperframe - timediff( &afterblit, &lastblit );
        if( timediff( &afterblit, &lastblit ) > ((usecperframe*2)+(usecperframe/2)) ) {
            fprintf( stderr, "viewcrawl: Frame dropped?\n" );
        }
        // fprintf( stderr, "diff: %d, error %d\n", timediff( &afterblit, &lastblit ), error );
        lastblit = curtime;
    }

    output->shutdown();
    return 0;
}

